From f829cd9a679003b14030679adbd08aae70cefc55 Mon Sep 17 00:00:00 2001
From: Amit Langote <amitlan@postgresql.org>
Date: Sun, 23 Feb 2025 17:19:28 +0900
Subject: [PATCH v1] Fix bug in cbc127917 to handle nested Append correctly

A sub-partitioned partition with a subplan that is an Append node was
not correctly reported in PlannedStmt.unprunableRelids. This omission
led to ExecGetRangeTableRelation() reporting an error when called from
CreatePartitionPruneState() to process its PartitionPruneInfo.

Reported-by: Alexander Lakhin <exclusion@gmail.com> (via sqlsmith)
Diagnosed-by: Tender Wang <tndrwang@gmail.com>
Discussion: https://postgr.es/m/74839af6-aadc-4f60-ae77-ae65f94bf607@gmail.com
---
 src/backend/executor/execPartition.c          | 14 +++++++++----
 src/backend/partitioning/partprune.c          |  8 +++++++-
 src/test/regress/expected/partition_prune.out | 20 +++++++++++++++++++
 src/test/regress/sql/partition_prune.sql      |  7 +++++++
 4 files changed, 44 insertions(+), 5 deletions(-)

diff --git a/src/backend/executor/execPartition.c b/src/backend/executor/execPartition.c
index 432eeaf9034..b86fc5ea297 100644
--- a/src/backend/executor/execPartition.c
+++ b/src/backend/executor/execPartition.c
@@ -2589,9 +2589,9 @@ ExecFindMatchingSubPlans(PartitionPruneState *prunestate,
  * find_matching_subplans_recurse
  *		Recursive worker function for ExecFindMatchingSubPlans
  *
- * Adds valid (non-prunable) subplan IDs to *validsubplans and the RT indexes
- * of their corresponding leaf partitions to *validsubplan_rtis if
- * it's non-NULL.
+ * Adds valid (non-prunable) subplan IDs to *validsubplans. If
+ * *validsubplan_rtis is non-NULL, it also adds the RT indexes of their
+ * corresponding partitions, but only if they are leaf partitions.
  */
 static void
 find_matching_subplans_recurse(PartitionPruningData *prunedata,
@@ -2628,7 +2628,13 @@ find_matching_subplans_recurse(PartitionPruningData *prunedata,
 		{
 			*validsubplans = bms_add_member(*validsubplans,
 											pprune->subplan_map[i]);
-			if (validsubplan_rtis)
+
+			/*
+			 * Subplan might be a nested Append / MergeAppend for a
+			 * sub-partitioned partition whose RT index need not be reported
+			 * to the caller.
+			 */
+			if (validsubplan_rtis && pprune->leafpart_rti_map[i])
 				*validsubplan_rtis = bms_add_member(*validsubplan_rtis,
 													pprune->leafpart_rti_map[i]);
 		}
diff --git a/src/backend/partitioning/partprune.c b/src/backend/partitioning/partprune.c
index ff926732f36..3faf3f8555c 100644
--- a/src/backend/partitioning/partprune.c
+++ b/src/backend/partitioning/partprune.c
@@ -687,7 +687,13 @@ make_partitionedrel_pruneinfo(PlannerInfo *root, RelOptInfo *parentrel,
 			if (subplanidx >= 0)
 			{
 				present_parts = bms_add_member(present_parts, i);
-				leafpart_rti_map[i] = (int) partrel->relid;
+
+				/*
+				 * Subplan might be a nested Append/MergeAppend for a
+				 * sub-partitioned partition.
+				 */
+				if (partrel->nparts == -1)
+					leafpart_rti_map[i] = (int) partrel->relid;
 
 				/* Record finding this subplan  */
 				subplansfound = bms_add_member(subplansfound, subplanidx);
diff --git a/src/test/regress/expected/partition_prune.out b/src/test/regress/expected/partition_prune.out
index 6f80b62a3b8..94a499fc8e0 100644
--- a/src/test/regress/expected/partition_prune.out
+++ b/src/test/regress/expected/partition_prune.out
@@ -4590,5 +4590,25 @@ table part_abc_view;
  2 | c | t
 (1 row)
 
+-- A case with nested Append with its own PartitionPruneInfo.
+create index on part_abc (a);
+create table part_abc_3 partition of part_abc for values in (3, 4) partition by list (a);
+create table part_abc_3_3 partition of part_abc_3 for values in (3);
+create table part_abc_3_4 partition of part_abc_3 for values in (4);
+explain (costs off) select min(a) over (partition by a order by a) from part_abc where a >= stable_one() + 1;
+                                      QUERY PLAN                                       
+---------------------------------------------------------------------------------------
+ WindowAgg
+   ->  Append
+         Subplans Removed: 1
+         ->  Index Only Scan using part_abc_2_a_idx on part_abc_2 part_abc_1
+               Index Cond: (a >= (stable_one() + 1))
+         ->  Append
+               ->  Index Only Scan using part_abc_3_3_a_idx on part_abc_3_3 part_abc_3
+                     Index Cond: (a >= (stable_one() + 1))
+               ->  Index Only Scan using part_abc_3_4_a_idx on part_abc_3_4 part_abc_4
+                     Index Cond: (a >= (stable_one() + 1))
+(10 rows)
+
 drop view part_abc_view;
 drop table part_abc;
diff --git a/src/test/regress/sql/partition_prune.sql b/src/test/regress/sql/partition_prune.sql
index 86621dcec0b..48273f8d027 100644
--- a/src/test/regress/sql/partition_prune.sql
+++ b/src/test/regress/sql/partition_prune.sql
@@ -1397,5 +1397,12 @@ using (select stable_one() + 2 as pid) as q join part_abc_1 pt1 on (q.pid = pt1.
 when matched then delete returning pt.a;
 table part_abc_view;
 
+-- A case with nested Append with its own PartitionPruneInfo.
+create index on part_abc (a);
+create table part_abc_3 partition of part_abc for values in (3, 4) partition by list (a);
+create table part_abc_3_3 partition of part_abc_3 for values in (3);
+create table part_abc_3_4 partition of part_abc_3 for values in (4);
+explain (costs off) select min(a) over (partition by a order by a) from part_abc where a >= stable_one() + 1;
+
 drop view part_abc_view;
 drop table part_abc;
-- 
2.43.0

